# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sys
import os
from twisted.application import service
from hep.services.datamanager import DataManagerService
from twisted.python import usage
from twisted.application import service
from yarn.util import loadClasses

services = loadClasses('hep.services', service.Service)

serviceOptions = []
for s in services:
    if hasattr(s, 'PORT'):
        serviceOptions.append(('%s-port' % s.NAME, None, str(s.PORT),
                               'Port to run %s service on' % s.NAME))

class Options(usage.Options):
    optParameters = [
        ('datadir', 'd', None, 'Directory to store Hep data files'),
        ('password', 'p', None, 'Administrator password'),
        ] + serviceOptions

    def postOptions(self):
        requiredOpts = ('datadir', 'password')
        for opt in requiredOpts:
            if not self.get(opt): raise usage.UsageError(
                "Missing required parameter '%s' (see --help for details)" % (
                opt))
        if not os.path.exists(self['datadir']):
            raise usage.UsageError(
                "The directory '%s' does not exist." % self['datadir'])
        if not os.path.isdir(self['datadir']):
            raise usage.UsageError(
                "The file '%s' is not a directory." % self['datadir'])
        for opt in self.keys():
            if opt.endswith('-port'):
                try:
                    int(self.get(opt))
                except ValueError, e:
                    raise usage.UsageError("--%s must be an integer" % opt)

def makeService(config):
    path = config.get('datadir')
    password = config.get('password')
    application = service.Application("Hep")
    dm = DataManagerService(os.path.abspath(path), password)
    dm.setName('Data Manager')
    dm.setServiceParent(application)

    for srvc in services:
        if srvc != DataManagerService:
            s = srvc()
            s.setName(s.NAME)
            if hasattr(s, 'PORT'):
                s.PORT = int(config.get('%s-port' % s.NAME))
                print '%s service on port %i' % (s.NAME, s.PORT)
            s.setServiceParent(application)

    return service.IServiceCollection(application)
