# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.cred import credentials, checkers, portal, error
#from twisted.internet import defer,app
from twisted.internet import defer
from twisted.application import service
from twisted.python import failure
import os.path

class Service(service.Service):
    NAME = 'Authenticator'
    
    def startService(self):
        realm = self.parent.getServiceNamed('Data Manager')
        self.portal = portal.Portal(realm)
        passwordFileName = os.path.join(realm.dataDir, 'passwd')
        self.authenticator = PasswordAuthenticator(
            realm.adminPass, passwordFileName)
        self.portal.registerChecker(self.authenticator)
        self.portal.registerChecker(checkers.AllowAnonymousAccess())

    def isValidUser(self, username):
        self.authenticator.refresh()
        print "isvaliduser '%s'?" % username
        return self.authenticator.passwords.has_key(username)

class PasswordAuthenticator:
    __implements__ = checkers.ICredentialsChecker
    credentialInterfaces = (credentials.IUsernamePassword,
                            credentials.IUsernameHashedPassword)
    
    def __init__(self, adminPassword, passwordFileName):
        self.adminPassword = adminPassword
        self.passwordFile = passwordFileName
        self.passwords = {}

    def refresh(self):
        self.passwords = {'admin': self.adminPassword}
        try:
            passwordFile = file(self.passwordFile)
        except:
            return
        for line in passwordFile:
            userName, password = line.strip().split(":")
            self.passwords[userName] = password
                
    def requestAvatarId(self, credentials):
        print 'AUTHENTICATOR: checking...', credentials.username
        self.refresh()
        passwd = self.passwords.get(credentials.username)
        if passwd is not None:
            return defer.maybeDeferred(credentials.checkPassword,
                                       passwd).addCallback(
                self._checkedPassword, credentials.username)
        else:
            return defer.fail(error.UnauthorizedLogin, "No such user")

    def _checkedPassword(self, matched, username):
        if matched:
            return username
        else:
            return failure.Failure(error.UnauthorizedLogin("Bad Password"))
    
