# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import urllib
import cPickle as pickle
import time, sha, urllib, os
from yarn import remote, url, message, auth, proxy
from twisted.internet import reactor, defer
from twisted.internet.defer import maybeDeferred
from twisted.application import service
from twisted.cred.portal import IRealm
from twisted.web import microdom

class DataManagerService(service.Service):
    NAME = 'Data Manager'
    __implements__ = IRealm

    def __init__(self, path, adminPass):
        self.dataDir = path
        self.adminPass = adminPass
        print "MY PATH IS", self.dataDir

    def startService(self):
        cacheDir = os.path.join(self.dataDir, "cache")
        if not os.path.exists(cacheDir): os.makedirs(cacheDir)
        self.connector = remote.Connector(cacheDir)
        self.connector.offline = 0
        
        self.avatarInterfaces = {}
        self.avatarInterfaces[HepUser] = lambda u: u
        
        # make sure dataDir has the sub-folders and files that it should
        userDir = os.path.join(self.dataDir, 'users')
        if not os.path.exists(userDir): os.makedirs(userDir)
        passwordFileName = os.path.join(self.dataDir, "passwd")
        if not os.path.isfile(passwordFileName):
            # create the file.
            open(passwordFileName, 'w').write('')

        self.users = {}
        for userName in os.listdir(os.path.join(self.dataDir, 'users')):
            self.loadUser(userName)
        self.users['anonymous'] = PublicUser(self.parent)
        # create the admin account if it's not there
        if not self.users.has_key('admin'): self.loadUser('admin')
        #reactor.callLater(0, self.loadConnections)

    def loadUser(self, userName):
        print "LOADING USER", userName
        userDir = os.path.join(self.dataDir, 'users', userName)
        self.users[userName] = HepUser(
            userName, userDir, self.connector, self.parent)

    def getUser(self, userName):
        if not userName: return
        if not self.users.has_key(userName):
            self.loadUser(userName)            
        return self.users[userName]

    def requestAvatar(self, avatarID, mind, *interfaces):
        for iface in interfaces:
            if iface in self.avatarInterfaces:
                if not avatarID:
                    user = self.users['anonymous']
                else:
                    user = self.getUser(avatarID)
                avatar = maybeDeferred(self.avatarInterfaces[iface], user)    
                avatar.addCallback(lambda a: (iface, a, lambda: None))
                return avatar
        else:
            raise NotImplementedError("no interface")

     
class HepUser:
    "Represents a user in Hep"    

    def __init__(self, name, baseFolder, connector, app):
        self.name = name
        self.baseFolder = baseFolder
        self.connector = connector
        self.app = app

        # make sure the user folder has the neccessary sub-folders
        new = False
        for path in ('messages',): # 'messages/connections', 'messages/index'):
            fullPath = os.path.join(*([baseFolder] + path.split('/')))
            if not os.path.exists(fullPath):
                new = True # a brand new user!
                print "HELLO NEW USER!"
                os.makedirs(fullPath)
        
        metaFile = os.path.join(baseFolder, 'meta.xml')
        keyFile = os.path.join(baseFolder, 'keychain.xml')
        self.keychain = auth.Keychain(keyFile)
        self.view = remote.Folder(remote.MetadataStore(metaFile))
        self.view.keychain = self.keychain
        self.config = UserConfig(os.path.join(self.baseFolder, 'config.ini'))
        messageFolder = os.path.join(baseFolder, 'messages')
        prox = HepDataProxy(messageFolder, connector)
        self.view.connectProxy(prox, "")
        print 'DM: refreshing'
        self.view.refresh().addCallback(self._connected, new)

    def _connected(self, result, new):
        # if this user is admin, post 'welcome' message
        if new and self.name == 'admin':
            print "POSTING WELCOME"
            m = message.Message()
            m['From'] = "Abe Fettig <abe@fettig.net>"
            m['To'] = "Hep Server Admin <admin@localhost>"
            m['Subject'] = "Welcome to Your New Hep Server"
            m.contents = """
Congratulations - your Hep server is up and running!

Hep is released under the GNU Public License (GPL).  This means that you are
free to install Hep on all the computers you own, give copies of Hep to your
friends, and even make changes to the underlying source code.  (You should
have recieved a copy of the GPL as part of your Hep download; you can also
read it online at http://www.gnu.org/copyleft/gpl.)

I believe that Hep is innovative and valuable software.  My decision to release
it under the GPL is a reflection of the fact that I want Hep to be as good
as I can possibly make it.  I hope you'll agree that the GPL licensing makes
Hep more valuable to you than it would be with a restrictive, proprietary
license.

If you find Hep useful, please consider making a donation to the project.  Your
donation will help reimburse me for the many hours I've put into developing
Hep, and allow me to continue working on making improvements.  You can make
a donation via the donations page on my website:

http://www.fettig.net/donations/

Thanks  for your support, and I hope you enjoy using Hep!

Abe
            """
            self.view.subFolders['Inbox'].postMessage(m)

    def urlToHash(self, u):
        "return hash of url"
        return sha.sha(str(u)).hexdigest()

    def getFolderFromHash(self, hash, _crawlStart=None):
        """
        return folder represented by hash, if it can be found in this user's
        sites, or None if no such folder can be found.
        """
        start = _crawlStart or self.view.subFolders['Sites']
        for f in start.subFolders.values():
            if self.urlToHash(f.url) == hash:
                return f
            recursiveLookup = self.getFolderFromHash(hash, f)
            if recursiveLookup: return recursiveLookup

    def getMessageForEmail(self, msg):
        "return copy of message with additional parts added and reply-to set"
        mCopy = msg.getCopy(generateMissingParts=1)
        mFrom = message.NameAndAddress(msg['From'])
        src = url.Url(msg['X-Yarn-Source-Url'])
        if not mFrom.address and src and src.protocol != 'file':
            mFrom.address = '%s@localhost' % self.urlToHash(src)
            mCopy['From'] = mFrom
        if not mCopy['Reply-To'] and self.config.get('weblogfolder'):
            hash = self.urlToHash(self.config['weblogfolder'])
            folder = self.getFolderFromHash(hash)
            if folder:
                replyTo = message.NameAndAddress()
                replyTo.name = folder.name
                replyTo.address = '%s@localhost' % hash
                mCopy['Reply-To'] = replyTo
                    
        return mCopy

from ConfigParser import RawConfigParser
class UserConfig(dict):
    "data storage for user config options"
    def __init__(self, filename):
        dict.__init__(self)
        self.filename = filename
        # read in file data
        if os.path.exists(self.filename):
            c = RawConfigParser()
            c.read(self.filename)
            for key, val in c.items('config'):
                self[key] = val
        
    def __setitem__(self, key, val):
        if type(key) != str or type(val) != str:
            raise TypeError, "Key and value must be strings"
        dict.__setitem__(self, key, val)
        self.save()

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.save()
    
    def save(self):
        outputLines = ['[config]']
        for key, val in self.items():
            outputLines.append('%s: %s' % (key, val))
        outputLines.append('') # end with newline
        file(self.filename, 'w').write("\n".join(outputLines))
        

class PublicUser:
    def __init__(self, app):
        self.app = app
        self.name = 'anonymous'


class HepDataProxy(proxy.Proxy):
    "proxy representing a user's local and remote messages."
    defaultFolders = ["Inbox"]
    
    def __init__(self, dir, connector):
        self.baseDir = dir
        proxy.Proxy.__init__(self, 'hep:', connector)
        for folder in self.defaultFolders:
            if not os.path.exists(os.path.join(self.baseDir, folder)):
                self.createFolder('', folder)

    def getFeatures(self, path):
        if path: raise KeyError, path
        return ["createfolder", "renamefolder", "deletefolder"]

    def getName(self, path):
        if path: raise KeyError, "No such path"
        return "Hep Message Folders"

    def getLinks(self, path):
        links = []
        opmlPath = os.path.join(self.baseDir, '..', 'sites.opml')
        if not os.path.exists(opmlPath):
            file(opmlPath, 'w+b').write(
                '''<?xml version="1.0" encoding="utf-8"?>
                <opml><head><title>My Sites</title></head>
                <body></body></opml>''')
        opmlFile = url.Link(
            'Sites',
            opmlPath,
            'child')
        links.append(opmlFile)
        for sub in os.listdir(self.baseDir):
            fullPath = os.path.join(self.baseDir, sub)
            links.append(url.Link(
                sub,
                fullPath,
                'child'))
        return defer.succeed(links)

    def createFolder(self, parent, newFolder):
        # FIXME: validate (existing folders, etc.), handle invalid characters.
        dir = os.path.join(self.baseDir, parent, newFolder)
        if not os.path.exists(dir): os.mkdir(dir)
        for subDir in ['new', 'cur', 'tmp']:
            subDirPath = os.path.join(dir, subDir)
            if not os.path.exists(subDirPath): os.mkdir(subDirPath)
        newLink = url.Link(newFolder, dir, 'child')
        return defer.succeed(newLink)
