# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
imap support for Hep. 
"""
import twisted.mail.imap4 as imap4 # local copy until next release of Twisted
from twisted.application import service
from twisted.internet.protocol import Factory
from twisted.internet import reactor, defer
from twisted.python import components
from twisted.python.util import InsensitiveDict
from yarn import message, events, features
from cStringIO import StringIO
import time

class Service(service.Service):
    PORT = 143
    NAME = 'imap'
    
    def startService(self):
        
        # add support for getting IMAP accounts out of the datamanager service
        dm = self.parent.getServiceNamed('Data Manager')
        dm.avatarInterfaces[imap4.IAccount] = openAccount
        
        portal = self.parent.getServiceNamed('Authenticator').portal
        self.factory = IMAPFactory(portal)
        self.factory.app = self.parent
        self.factory.protocol = HepIMAPServer
        
        reactor.listenTCP(self.PORT, self.factory)

class IMAPFactory(Factory):
    
    def __init__(self, portal):
        self.portal = portal
    
    def buildProtocol(self, address):
        p = self.protocol()
        p.portal = self.portal
        p.factory = self
        return p

class HepIMAPServer(imap4.IMAP4Server):
    pass
    
def openAccount(user):
    print "OPENING ACCOUNT FOR", user
    return HepAccount(user)

class HepAccount(components.Adapter):
    __implements = (imap4.IAccount,)
            
    def listMailboxes(self, ref, wildcard):
        print "IMAP: listMapboxes", ref, wildcard
        folder = self._getFolder(ref)
        boxes = []
        self._loadBoxes(ref, folder, boxes)
        return boxes

    def _loadBoxes(self, path, folder, boxList):
        for key, box in folder.subFolders.items():
            fullPath = key #.encode('imap4-utf-7')
            if path:
                fullPath = "/".join((path, fullPath))
            boxList.append((fullPath, HepMailbox(box, self.original)))
            self._loadBoxes(fullPath, box, boxList)

    def _getFolder(self, path):
        print "_getFolder", repr(path)
        f = self.original.view
        parts = path.split('/')
        while parts:
            part = parts.pop(0) #.decode('imap4-utf-7')
            if part.lower() == 'inbox': part = 'Inbox'
            if not part: continue
            f = f.subFolders[part]
        return f
        
    def create(self, path):
        if not path.count('/'):
            return self.original.view.createFolder(path)
        else:
            raise imap4.MailboxException("Permission denied")

    def delete(self, path):
        raise imap4.MailboxException("Permission denied.")
    
    def isSubscribed(self, path):
        return self._getFolder(path).metadata.get('imap.subscribed', 0)
        
    def subscribe(self, path):
        print "IMAP: subscribe", path
        f = self._getFolder(path)
        f.metadata['imap.subscribed'] = 1
        f.saveSoon()
        return 1
        
    def unsubscribe(self, path):
        print "IMAP: unsubscribe", path
        f = self._getFolder(path)
        f.metadata['imap.subscribed'] = 0
        f.saveSoon()
        return 1
        
    def select(self, path, rw=True):
        folder = self._getFolder(path)
        return HepMailbox(folder, self.original)
        

flagMap = {
    '\Seen': ('seen', 1),
    '\Unseen': ('seen', 0),
    '\Deleted': ('deleted', 1)
    }
    
class HepMailbox(object):
    __implements__ = (imap4.IMailbox,)

    def __init__(self, folder, user):
        self.folder = folder
        self.url = self.folder.url
        self.user = user
        print "IMAP: HepMailbox: My PATH IS", repr(self.folder.path)
        # add imap fields to the view if they're not there already.
        defaults = {'imap.counter': 1,
                    'imap.uidvalidity': int(time.time()),
                    }
        for field, val in defaults.items():
            self.folder.metadata.setdefault(field, val)
        self.folder.saveSoon()
        self.uidMap = {}
        self.listeners = []
        self.folder.setEventHandler(events.MESSAGE_POSTED, self.gotNewMessage)

    def buildUidMap(self):
        d = self.folder.listMessages()
        d.addCallback(self._loadedMessageList)
        return d
    
    def _loadedMessageList(self, messageKeys):
        for key in messageKeys:
            self._addMessageKey(key)
        self.folder.saveSoon()
        return 1

    def _addMessageKey(self, key):
        self.folder.messageMetadata.setdefault(key, {})
        messageData = self.folder.messageMetadata[key]
        if not messageData.has_key('imap.uid'):
            messageData['imap.uid'] = self.folder.metadata['imap.counter']
            self.folder.metadata['imap.counter']+=1
        self.uidMap[messageData['imap.uid']] = key
        return messageData['imap.uid']

    def getHierarchicalDelimiter(self):
        print "MBOX: getHier..."
        return '/'
    
    def getFlags(self):
        print "MBOX: getFlags", repr(self.folder.path)
        return flagMap.keys()
        
    def getMessageCount(self):
        print "MBOX: getMessageCount", repr(self.folder.name)
        return self.folder.summary.count
        return self.folder.summary.refresh().addCallback(
            (lambda _, s: s.count), self.folder.summary)

    def getRecentCount(self):
        print "MBOX: getRecentCount", repr(self.folder.name)
        return 0

    def getUnseenCount(self):
        print "MBOX: getUnseenCOunt", repr(self.folder.name)
        return self.folder.summary.unseenCount
        return self.folder.summary.refresh().addCallback(
            (lambda _, s: s.count), self.folder.unseenCount)

    def isWriteable(self):
        print "MBOX: isWriteable",features.POST_MESSAGE in self.folder.features
        return features.POST_MESSAGE in self.folder.features

    def getUIDValidity(self):
        print "MBOX: getUIDValidity", repr(self.folder.name)
        return self.folder.metadata['imap.uidvalidity']
        
    def getUID(self, messageNo):
        print "MBOX: getUID"
        uids = self.uidMap.keys()
        uids.sort()
        return uids[messageNo-1]
        
    def getUIDNext(self):
        print "MBOX: getUIDNext", repr(self.folder.name)
        return self.folder.metadata['imap.counter']   
                
    def mapMessageIds(self, messageSet, uid):
        "return a dict of {messageKey: messageSetID} for the given messageSet."
        if not messageSet.last:
            if uid:
                if not self.uidMap:
                    messageSet.last = 0
                else:
                    messageSet.last = max(self.uidMap.keys())
            else:
                messageSet.last = len(self.uidMap)
        keys = {}
        for m in messageSet:
            if not uid:
                keys[self.uidMap[self.getUID(m)]] = m
            elif self.uidMap.has_key(m):
                keys[self.uidMap[m]] = m
            else:
                print "IMAP: KEY ERROR! NO SUCH MESSAGE", m
        return keys

    def fetch(self, messages, uid):
        print "MBOX: fetch", messages, uid, repr(self.folder.path)
        finished = defer.Deferred()
        if not self.uidMap:
            self.buildUidMap().addCallback(
                self._finishFetching, messages, uid, finished)
        else:
            reactor.callLater(
                0, self._finishFetching, None, messages, uid, finished)
        return finished 

    def _finishFetching(self, _, messages, uid, finished):
        idsToFetch = self.mapMessageIds(messages, uid)
        fetchers = []
        for key, mUID in idsToFetch.items():
            d = self.folder.getMessage(key)
            # Since there's no way to see if they called BODY.PEEK or whatever,
            # assume that fetching a single message should mark it as read.
            # FIXME: this is a hack!
            #if len(idsToFetch) == 1:
            #    d.addCallback(self._markMessageAsSeen, key, mUID)
            d.addCallback(self._wrapMessage, key, mUID)
            fetchers.append(d)
        return defer.DeferredList(fetchers).addCallback(
            self._fetched, finished)

    def _fetched(self, results, finished):
        messages = iter([vals for succeeded, vals in results if succeeded])
        finished.callback(messages)

    def _markMessageAsSeen(self, m, key, uid=None):
        print "IMAP: marking Message as seen:", key
        self.folder.messageMetadata[key]['seen'] = 1
        self.folder.saveSoon()
        return m
    
    def _wrapMessage(self, m, key, uid):
        return uid, MessageAdapter(self.folder, key,
                                   self.user.getMessageForEmail(m).mimeMessage)

    def addListener(self, listener):
        print "MBOX: addListener"
        self.listeners.append(listener)
        return 1
    
    def removeListener(self, listener):
        print "MBOX: removeListener"
        self.listeners.remove(listener)
        return 1
        
    def requestStatus(self, path):
        try:
            print "REQUEST STATUS"
            return imap4.statusRequestHelper(self, path)
        except Exception, e:
            print "REQUEST STATUS ERROR"
            import traceback
            traceback.print_exc()
            
            
    
    #def search(self, query, uid):
    #    print "MBOX: Search"
    #    return []

    def addMessage(self, msg, flags=(), date=None):
        print "MBOX: addMessage", flags, date
        m = message.Message(msg)
        if date: m.setTimestamp(date)
        return self.folder.postMessage(m).addCallback(
            self._postedMessage, flags)

    def _postedMessage(self, key, flags):
        for flag in flags:
            if flagMap.has_key(flag):
                f, val = flagMap[flag]
                self.folder.messageMetadata[key][f] = val
        return str(self._addMessageKey(key))

    def store(self, messages, flags, mode, uid):
        print "MBOX: store", messages, flags, mode, uid
        keyMap = self.mapMessageIds(messages, uid)
        results = {}
        for key, uid in keyMap.items():
            results[uid] = []
            for flag in flags:
                if flagMap.has_key(flag):
                    f, val = flagMap.get(flag)
                    if mode == 1:
                        self.folder.messageMetadata[key][f] = val
                    else:
                        if self.folder.messageMetadata[key].has_key(f):
                            del(self.folder.messageMetadata[key][f])
                results[uid].append(flag)
        self.folder.saveSoon()
        return results

    def expunge(self):
        print "MBOX: expunge()"
        return []

    def destroy(self):
        print "MBOX: destroy"
        return 0

    def gotNewMessage(self, key):
        print "MBOX: GOT NEW MESSAGE!"
        for listener in self.listeners:
            listener.newMessages(self.folder.summary.count,
                                 1) # FIXME... should be \Recent count
        self.buildUidMap()
    

class MessagePartAdapter:
    __implements__ = (imap4.IMessagePart)

    def __init__(self, mimeMessage):
        self.message = mimeMessage

    def getHeaders(self, negate, *names):
        print "MESSAGEADAPTER: getHeaders"
        if not names: names = self.message.keys()
        headers = {}
        if negate:
            for header in self.message.keys():
                if header.upper() not in names:
                    headers[header.lower()] = str(self.message.get(header, ''))
        else:
            for name in names:
                headers[name.lower()] = str(self.message.get(name, ''))
        print "HEADERS", headers
        return headers
        
    def getInternalDate(self):
        print "RETURNING DATE"
        return self.message['Date']
        
    def getBodyFile(self):
        print "MSG: GETTING BODY FILE"        
        # this should return the unencoded body, but it doesn't:
        #body = self.message.get_payload()
        # so I'm using this fallback hack
        body = str(self.message).split('\n\n', 1)[1]
        return StringIO(body)

    def getSize(self):
        print "RETURNING SIZE",
        size = len(str(self.message))
        print size
        return size
        
    def isMultipart(self):
        return self.message.is_multipart()

    def getSubPart(self, partNo):
        print "MSG: getSubPart"
        return MessagePartAdapter(self.message.get_payload(partNo))

class MessageAdapter(MessagePartAdapter):
    __implements__ = (imap4.IMessage)
    
    def __init__(self, folder, key, message):
        self.metadata = folder.messageMetadata[key]
        MessagePartAdapter.__init__(self, message)
    
    def getUID(self):
        print 'RETURNING UID'
        return self.metadata['imap.uid']

    def getFlags(self):
        flags = []
        for imapFlag, (hepFlag, val) in flagMap.items():
            if self.metadata.get(hepFlag, 0) == val:
                flags.append(imapFlag)
        print "getFlags: RETURNING FLAGS:", flags
        return flags      
