# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
SMTP support for Hep
"""
from twisted.application import service
from twisted.internet.protocol import Factory
from twisted.mail import smtp
from twisted.internet import reactor, defer
from twisted.cred import credentials,portal
from yarn import message, url, remote, features

class SMTPService(service.Service):
    PORT = 25
    NAME = 'smtp'
    
    def startService(self):
        # register support for SMTP user wrapper with the datamanager service
        dm = self.parent.getServiceNamed('Data Manager')
        dm.avatarInterfaces[smtp.IMessageDelivery] = HepUserDelivery
#        portal = self.parent.getServiceNamed('Authenticator').portal
        self.factory = SMTPFactory(portal)
        self.factory.app = self.parent
        reactor.listenTCP(self.PORT, self.factory)


class MessagePoster(object):
    __implements__ = [smtp.IMessage]

    def __init__(self, folder):
        self.folder = folder
        self.lines =[]

    def lineReceived(self, line):
        self.lines.append(line)

    def eomReceived(self):
        m = message.Message("\n".join(self.lines))
        return self.folder.postMessage(m)

    def connectionLost(self):
        del(self.lines)
        self.lines = []


class HepUserDelivery(object):
    __implements__ = [smtp.IMessageDelivery]

    def __init__(self, user):
        self.user = user

    def receivedHeader(self, *args):
        # FIXME: implement ;)
        return 'Recieved: from somewhere by localhost;'

    def validateFrom(self, helo, origin):
        return origin

    def validateTo(self, user):
        # FIXME: validate domain
        username = user.dest.local
        if self.user.app.getServiceNamed(
            'Authenticator').isValidUser(username):
            dm = self.user.app.getServiceNamed('Data Manager')
            inbox = dm.getUser(username).view.subFolders['Inbox']
            return lambda: MessagePoster(inbox)
        else:
            if self.user.name == 'anonymous':
                print 'user is anonymous'
                raise smtp.SMTPBadRcpt(user)
            else:
                f = self.user.getFolderFromHash(user.dest.local)
                if f:
                    return lambda: MessagePoster(f)
                else:
                    raise smtp.SMTPBadRcpt(user, resp='No such site')

    def getFolderFromUrl(self, u):
        f = remote.Folder(keychain=self.user.keychain)
        return f.connect(u).addCallback(self._connectedToFolder, f).addErrback(
            self._connectFailed, f)
        
    def _connectedToFolder(self, result, folder):
        if features.POST_MESSAGE in folder.features:
            return lambda: MessagePoster(folder)
        else:
            raise smtp.SMTPBadRcpt('"%s"@%s' % (folder.url, 'localhost'),
                                   resp="can't post message to that site.")

    def _connectFailed(self, fail, folder):
        raise smtp.SMTPBadRcpt('"%s"@%s' % (folder.url, 'localhost'),
                               resp=fail.getErrorMessage())
                   

class SMTPFactory(Factory):
    def __init__(self, portal):
        self.portal = portal
    
    def buildProtocol(self, address):
        p = smtp.ESMTP({'CRAM-MD5': credentials.CramMD5Credentials,
                        'PLAIN': credentials.UsernamePassword})
        p.portal = self.portal
        p.factory = self
        return p
